import { async as subject } from 'most-subject';
import { UniqGuard } from '../uniq';
import { createIDType, createSeq } from '../id';
import { effectFabric, eventFabric } from './fabric';
import { storeInjector } from '../inject-reducer';
var nextID = createIDType();
export class Store {
  constructor() {
    this.uniq = new UniqGuard();
    this.injector = storeInjector();
    this.id = nextID();
    this.seq = createSeq();
    this.update$ = subject();
    this.state$ = this.update$.map(({
      state
    }) => state).multicast();

    this.dispatch$ = (() => {
      var subj = subject();
      subj.skipRepeats().observe(e => {
        if (isAction(e)) this.dispatch(e);
      });
      return subj;
    })();
  }

  *scope() {
    yield* this.scopeName;
  }

  getState() {
    return this.state$;
  }

  mergeEvents(emitter) {
    // emitter.event$ = subject()
    // emitter.event$.observe(e => this.dispatch$.next(e))
    emitter.event$ = this.dispatch$;
  }

  event(description) {
    var result = eventFabric(description, this); // this.mergeEvents(result)

    return result;
  }

  effect(description) {
    var result = effectFabric(description, this); // this.mergeEvents(result)

    return result;
  }

  subdomain(name) {
    var result = new Store();
    result.dispatch = this.dispatch;
    result.stateGetter = this.stateGetter;
    result.reduxSubscribe = this.reduxSubscribe;
    result.dispatch$ = this.dispatch$;
    result.scopeName = [...this.scopeName, name];
    result.injector = this.injector;
    result.getState = this.stateGetter;
    result.subscribe = this.reduxSubscribe;
    result.replaceReducer = this.replaceReducer;
    return result;
  }

}
export function hideProperties(store) {
  var {
    dispatch$,
    update$,
    state$
  } = store;
  Object.defineProperties(store, {
    dispatch$: Object.assign({
      value: dispatch$
    }, nonEnumProp),
    update$: Object.assign({
      value: update$
    }, nonEnumProp),
    state$: Object.assign({
      value: state$
    }, nonEnumProp)
  });
  return store;
}
var nonEnumProp = {
  writable: true,
  enumerable: false,
  configurable: true
};

function isAction(value)
/*::%checks*/
{
  return typeof value === 'object' && value != null && typeof value.type === 'string' && value.type.length > 0;
}
//# sourceMappingURL=store.js.map