"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Event = Event;

var _most = require("most");

var _id = require("../id");

var _register = require("./register");

//@noflow
// import type {Subject} from 'most-subject'
// import type {Named, Typed, WithStateLink, Emitter} from '../index.h'
// import {carrier, type Carrier} from './carrier'
var nextID = (0, _id.createIDType)(); // import {registerType} from './register'
// import {on} from './pubsub'
// import {Act, type ActTag} from './act'

function Event(description, actionConstructor) {
  this.id = nextID();
  this.used = 0;
  var isSerializable = typeof description === 'string' && /^[0-9A-Z_]+$/.test(description);

  if (isSerializable) {
    (0, _register.check)(description);
    (0, _register.add)(description);
  } else {
    ++id;
  }

  this.actionType = description;
  this.typeId = id;
  this.actionConstructor = actionConstructor;

  function event(payload) {
    return event.run(payload);
  }

  Object.setPrototypeOf(event, Event.prototype);
  Object.assign(event, this);
  return event;
}

Event.prototype.run = function run(payload) {
  this.used += 1;
  return this.actionConstructor(this.typeId, this.getType(), payload, e => this.emit(e));
};

Event.prototype.getType = function getType() {
  return [...this.scope(), this.actionType].join('/');
};

Event.prototype.emit = function emit(value) {
  this.event$.next(value);
};

Event.prototype.apply = function apply(that, args) {
  return this.run(args[0]);
};

Event.prototype.watch = function watch(handler) {
  var result = this.epic(data$ => watcher(data$, handler)); // result.observe(e => this.emit(e))

  return result;
};

var id = 0;

function watcher(data$, handler) {
  return data$.map(async ({
    data,
    state
  }) => {
    try {
      var result = await Promise.resolve(handler(data, state));
      return result;
    } catch (error) {
      console.warn(`\n  Watcher throw\n  ${error && error.message || ''}`);
    }
  }).map(_most.fromPromise).join();
}
//# sourceMappingURL=event.js.map