"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createReducer = createReducer;
exports.ReducerObject = exports.ReducerKeyed = exports.ReducerRoot = exports.ReducerAbstract = void 0;

var _immutable = require("immutable");

var _id = require("./id");

var nextID = (0, _id.createIDType)();

class ReducerAbstract {
  constructor() {
    this.update = next => {
      this.seq = (0, _id.incSeq)(this.seq);
      this.propagation(next);
    };
  }
  /*::+*/
  // clone(): ReducerAbstract<State, Parent> {
  //   const result: ReducerAbstract<State, Parent> = new ReducerAbstract
  //   Object.assign(
  //     result,
  //     shape(this)
  //   )
  //   result.instance = incSeq(result.instance)
  //   return result
  // }


}

exports.ReducerAbstract = ReducerAbstract;

class ReducerRoot extends ReducerAbstract {
  constructor(...args) {
    var _temp;

    return _temp = super(...args), this.setter = (_, value) => {
      value;
    }, _temp;
  }

}

exports.ReducerRoot = ReducerRoot;

class ReducerKeyed extends ReducerAbstract {
  getter(parent) {
    return parent[this.key];
  }

  setter(parent, state) {
    return Object.assign({}, parent, {
      [this.key]: state
    });
  }

}

exports.ReducerKeyed = ReducerKeyed;

class ReducerObject extends ReducerAbstract {}

exports.ReducerObject = ReducerObject;

var gt = () => new ReducerKeyed();

function shape(reducer) {
  var {
    id,
    instance,
    seq,
    events,
    defaultState,
    propagation
  } = reducer;
  return {
    id,
    instance,
    seq,
    events,
    defaultState,
    propagation
  };
}

function createReducer(defaultState) {
  var result = new ReducerAbstract();
  var shape = {
    id: nextID(),
    instance: (0, _id.createSeq)(),
    seq: (0, _id.createSeq)(),
    events: (0, _immutable.Set)(),
    defaultState
  };
  Object.assign(result, shape);
  return result;
} // export class ReduxReducer<State = mixed> {
//   id: ID = nextID()
//   seq: SEQ = createSeq()
//   defaultState: State
//   types: Map<string, Function> = Map()
//   constructor(defaultState: State) {
//     this.defaultState = defaultState
//     //$off
//     this.on = this.on.bind(this)
//     //$off
//     this.off = this.off.bind(this)
//   }
//   ensureState(state?: State): State {
//     if (state === undefined) return this.defaultState
//     return state
//   }
//   reduce<P, M>(state: State, message: {
//     type: string,
//     payload: P,
//     meta: M,
//   }): State {
//     const handler = this.types.get(message.type)
//     const trueState = this.ensureState(state)
//     if (message && handler) {
//       const result = handler(trueState, message.payload, message.meta)
//       if (result !== undefined) return result
//     }
//     return trueState
//   }
//   on<P, A/*::: Event<P, any, State> | $ReadOnlyArray<Event<P, any, State>>*/>(
//     event: A,
//     handler: (state: State, payload: P) => State
//   ): this {
//     if (Array.isArray(event)) {
//       event.forEach(message => {
//         this.on(message, handler)
//       })
//     } else {
//       this.types.set(event, handler)
//     }
//     return this
//   }
//   off(event: Event<*, any, State> | $ReadOnlyArray<Event<*, any, State>>) {
//     if (Array.isArray(event)) {
//       event.forEach(this.off)
//     } else {
//       this.types.delete(event)
//     }
//     return this
//   }
//   manual(fn: (on: $PropertyType<this, 'on'>, off: $PropertyType<this, 'off'>) => void) {
//     fn(this.on, this.off)
//     return this
//   }
// }
//# sourceMappingURL=reducer.js.map