"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.eventFabric = eventFabric;
exports.effectFabric = effectFabric;

var _event = require("../carrier/event");

var _carrier = require("../carrier/carrier");

var _effect = require("../carrier/effect");

function reassign(msg) {
  function messageCarrier(payload) {
    return this.run(payload);
  }

  var actionBind = messageCarrier.bind(msg);
  Object.setPrototypeOf(actionBind, msg);
  Object.assign(actionBind, msg);
  return msg;
}

function reassignEffect(msg) {
  function effectCarrier(payload) {
    return this.run(payload);
  }

  var actionBind = effectCarrier.bind(msg);
  Object.setPrototypeOf(actionBind, _effect.Effect.prototype);
  Object.assign(actionBind, msg); // actionBind.use = msg.use.bind(msg)

  return msg;
}

function eventFabric(description, store) {
  var {
    update$
  } = store;
  var result = new _event.Event(description, _carrier.carrier);

  result.scope = () => store.scope();

  var epic$ = update$.filter(({
    data
  }) => data.typeId === result.typeId).skipRepeats().multicast();
  var payload$ = epic$.map(({
    data,
    state
  }) => ({
    data: data.payload,
    state
  })).skipRepeats().tap(console.log.bind(console)).multicast();
  result.epic$ = epic$;

  result.getState = () => store.getState();

  store.mergeEvents(result);

  result.epic = function (epic) {
    var epic$ = epic(payload$, result.getState());
    epic$.observe(resultObserver);
    return epic$;

    function resultObserver(value) {
      if (value instanceof _carrier.Carrier) return result.emit(value);
      if (typeof value === 'object' && value != null && typeof value.type === 'string') return result.emit(value);
    }
  };

  return reassign(result);
}

function effectFabric(description, store) {
  var {
    update$
  } = store;

  function effectInstance(payload) {
    return effectInstance.run(payload);
  }

  Object.setPrototypeOf(effectInstance, _effect.Effect.prototype);

  _effect.Effect.call(effectInstance, description);

  var result = effectInstance;

  result.scope = () => store.scope();

  result.done = eventFabric(`${description} done`, store);
  result.fail = eventFabric(`${description} fail`, store);
  var epic$ = update$.filter(({
    data
  }) => data.typeId === result.typeId).map(e => e).multicast();
  var payload$ = epic$.map(({
    data,
    state
  }) => ({
    data: data.payload,
    state
  })).multicast();
  result.epic$ = epic$;

  result.getState = () => store.getState();

  store.mergeEvents(result);

  result.epic = function (epic) {
    var epic$ = epic(payload$, result.getState());
    epic$.observe(resultObserver);
    return epic$;

    function resultObserver(value) {
      if (value instanceof _carrier.Carrier) return result.emit(value);
      if (typeof value === 'object' && value != null && typeof value.type === 'string') return result.emit(value);
    }
  };

  return reassignEffect(result);
}
//# sourceMappingURL=fabric.js.map