"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.hideProperties = hideProperties;
exports.Store = void 0;

var _mostSubject = require("most-subject");

var _uniq = require("../uniq");

var _id = require("../id");

var _fabric = require("./fabric");

var _injectReducer = require("../inject-reducer");

var nextID = (0, _id.createIDType)();

class Store {
  constructor() {
    this.uniq = new _uniq.UniqGuard();
    this.injector = (0, _injectReducer.storeInjector)();
    this.id = nextID();
    this.seq = (0, _id.createSeq)();
    this.update$ = (0, _mostSubject.async)();
    this.state$ = this.update$.map(({
      state
    }) => state).multicast();

    this.dispatch$ = (() => {
      var subj = (0, _mostSubject.async)();
      subj.skipRepeats().observe(e => {
        if (isAction(e)) this.dispatch(e);
      });
      return subj;
    })();
  }

  *scope() {
    yield* this.scopeName;
  }

  getState() {
    return this.state$;
  }

  mergeEvents(emitter) {
    // emitter.event$ = subject()
    // emitter.event$.observe(e => this.dispatch$.next(e))
    emitter.event$ = this.dispatch$;
  }

  event(description) {
    var result = (0, _fabric.eventFabric)(description, this); // this.mergeEvents(result)

    return result;
  }

  effect(description) {
    var result = (0, _fabric.effectFabric)(description, this); // this.mergeEvents(result)

    return result;
  }

  subdomain(name) {
    var result = new Store();
    result.dispatch = this.dispatch;
    result.stateGetter = this.stateGetter;
    result.reduxSubscribe = this.reduxSubscribe;
    result.dispatch$ = this.dispatch$;
    result.scopeName = [...this.scopeName, name];
    result.injector = this.injector;
    result.getState = this.stateGetter;
    result.subscribe = this.reduxSubscribe;
    result.replaceReducer = this.replaceReducer;
    return result;
  }

}

exports.Store = Store;

function hideProperties(store) {
  var {
    dispatch$,
    update$,
    state$
  } = store;
  Object.defineProperties(store, {
    dispatch$: Object.assign({
      value: dispatch$
    }, nonEnumProp),
    update$: Object.assign({
      value: update$
    }, nonEnumProp),
    state$: Object.assign({
      value: state$
    }, nonEnumProp)
  });
  return store;
}

var nonEnumProp = {
  writable: true,
  enumerable: false,
  configurable: true
};

function isAction(value)
/*::%checks*/
{
  return typeof value === 'object' && value != null && typeof value.type === 'string' && value.type.length > 0;
}
//# sourceMappingURL=store.js.map