//@flow

import {List, Map, Set} from 'immutable'

import type {Event} from './carrier/event'
import {type ID, createIDType, type SEQ, createSeq, incSeq} from './id'

const nextID = createIDType()

interface ReducerShape<State = void> {
  id: ID,
  instance: SEQ,
  seq: SEQ,
  events: Set<number>,
  defaultState: State,
  propagation: (next: State) => void,
}

export class ReducerAbstract<State, Parent> implements ReducerShape<State> {
  id: ID
  instance: SEQ
  seq: SEQ
  events: Set<number>
  defaultState: State
  propagation: (next: State) => void
  update = (next: State) => {
    this.seq = incSeq(this.seq)
    this.propagation(next)
  }
  /*::+*/getter: (parent: Parent) => State
  /*::+*/setter: (parent: Parent, value: State) => Parent
  // clone(): ReducerAbstract<State, Parent> {
  //   const result: ReducerAbstract<State, Parent> = new ReducerAbstract
  //   Object.assign(
  //     result,
  //     shape(this)
  //   )
  //   result.instance = incSeq(result.instance)
  //   return result
  // }
}

export class ReducerRoot<State> extends ReducerAbstract<State, void> {
  setter: * = (_, value) => {
    value
  }
}

export class ReducerKeyed<
  Parent/*:{[string]: any}*/,
  Key/*:$Keys<Parent>*/,
  State = $ElementType<Parent, Key>,
> extends ReducerAbstract<State, Parent> {
  key: Key
  getter(parent: Parent): State {
    return parent[this.key]
  }
  setter(parent: Parent, state: State): Parent {
    return {...parent, [this.key]: state}
  }
}

export class ReducerObject<State/*:{[string]: any}*/, Parent> extends ReducerAbstract<State, Parent> {

}


const gt = (): ReducerKeyed<{bar: 'foo', value: 'test'}, 'bar'> => new ReducerKeyed

function shape<State>(reducer: ReducerAbstract<State, any>): ReducerShape<State> {
  const {
    id, instance, seq,
    events, defaultState, propagation,
  } = reducer
  return {
    id, instance, seq,
    events, defaultState, propagation,
  }
}

export function createReducer<State>(
  defaultState: State
): ReducerAbstract<State, *> {
  const result = new ReducerAbstract
  const shape = {
    id: nextID(),
    instance: createSeq(),
    seq: createSeq(),
    events: Set(),
    defaultState,
  }
  Object.assign(result, shape)
  return result
}

// export class ReduxReducer<State = mixed> {
//   id: ID = nextID()
//   seq: SEQ = createSeq()
//   defaultState: State
//   types: Map<string, Function> = Map()
//   constructor(defaultState: State) {
//     this.defaultState = defaultState
//     //$off
//     this.on = this.on.bind(this)
//     //$off
//     this.off = this.off.bind(this)
//   }
//   ensureState(state?: State): State {
//     if (state === undefined) return this.defaultState
//     return state
//   }
//   reduce<P, M>(state: State, message: {
//     type: string,
//     payload: P,
//     meta: M,
//   }): State {
//     const handler = this.types.get(message.type)
//     const trueState = this.ensureState(state)
//     if (message && handler) {
//       const result = handler(trueState, message.payload, message.meta)
//       if (result !== undefined) return result
//     }
//     return trueState
//   }

//   on<P, A/*::: Event<P, any, State> | $ReadOnlyArray<Event<P, any, State>>*/>(
//     event: A,
//     handler: (state: State, payload: P) => State
//   ): this {
//     if (Array.isArray(event)) {
//       event.forEach(message => {
//         this.on(message, handler)
//       })
//     } else {
//       this.types.set(event, handler)
//     }

//     return this
//   }

//   off(event: Event<*, any, State> | $ReadOnlyArray<Event<*, any, State>>) {
//     if (Array.isArray(event)) {
//       event.forEach(this.off)
//     } else {
//       this.types.delete(event)
//     }
//     return this
//   }
//   manual(fn: (on: $PropertyType<this, 'on'>, off: $PropertyType<this, 'off'>) => void) {
//     fn(this.on, this.off)
//     return this
//   }

// }
